//Project: CNC T8
//Homepage: www.HomoFaciens.de
//Author Norbert Heinz
//Version: 0.1
//Creation date: 21.03.2017
//This program is free software you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation version 3 of the License.
//This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//For a copy of the GNU General Public License see http://www.gnu.org/licenses/
//
//compile with gcc commands-CNC.c -o commands-CNC -lm
//For details see:
//http://www.HomoFaciens.de

#include <stdio.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <dirent.h>
#include <math.h>
#include <unistd.h>
#include <fcntl.h>
#include <complex.h>



#define BAUDRATE B115200
#define ARDUINOPORT "/dev/ttyUSB0"
//#define ARDUINOPORT "/dev/ttyACM0"
#define FALSE 0
#define TRUE 1

//Number of bytes to sent via USB before waiting for 'r' from Arduino
//Must be
#define MAX_ARDUINO_BUFFER 1
#define MAX_FEED_RATE     320

#define PI 3.1415927

int KeyHit = 0;
int KeyCode[5];
int SingleKey=0;

FILE *GcodeFile;

int  MaxRows = 24;
int  MaxFileRows = 10;
int  FilesFound = 0;
int  MaxCols = 80;
int  MessageX = 1;
int  MessageY = 24;
double Scale = 1.0;
double OldScale = 1.0;
double MoveLength = 1;
double xMin = 1000000, xMax = -1000000;
double yMin = 1000000, yMax = -1000000;
double zMin = 1000000, zMax = -1000000;
char FileName[1000] = "";
double currentX = 0.0, currentY = 0.0, currentZ = 0.0, currentFeedRate = MAX_FEED_RATE, latestFeedRate = MAX_FEED_RATE;

double newX, newY, newZ, zStart = 50.0, feedRateNew;

int stopPlot = 0;


int SpindleOn = 0;
int CoolantOn = 0;
int PauseTime = 0;
char LastGCodeCharacter = 0;
int LastGcodeNumber = 0;
int MaxSpeed = 1;
int AbsolutePositioning = 1;

double StepsPermmX = 1200.0 / 10.00;
double StepsPermmY = 1200.0 / 10.00;
double StepsPermmZ = 1200.0 / 10.00;
long zHubUp = 0, zHubDown = -100;

int fd = 0;
struct termios TermOpt;
int waitForKeyHit = 1;

int relayOn = 0;
int coolantOn = 0;
int OldStepPause = 0;
int StepPause = 100;
int StepPauseArduino = 1;

int plotterSteps = 0;

int arduinoBuffer = 0;


char PicturePath[1000];




//+++++++++++++++++++++++ Start gotoxy ++++++++++++++++++++++++++
//Thanks to 'Stack Overflow', found on http://www.daniweb.com/software-development/c/code/216326
int gotoxy(int x, int y) {
  char essq[100]; // String variable to hold the escape sequence
  char xstr[100]; // Strings to hold the x and y coordinates
  char ystr[100]; // Escape sequences must be built with characters

  //Convert the screen coordinates to strings.
  sprintf(xstr, "%d", x);
  sprintf(ystr, "%d", y);

  //Build the escape sequence (vertical move).
  essq[0] = '\0';
  strcat(essq, "\033[");
  strcat(essq, ystr);

  //Described in man terminfo as vpa=\E[%p1%dd. Vertical position absolute.
  strcat(essq, "d");

  //Horizontal move. Horizontal position absolute
  strcat(essq, "\033[");
  strcat(essq, xstr);
  // Described in man terminfo as hpa=\E[%p1%dG
  strcat(essq, "G");

  //Execute the escape sequence. This will move the cursor to x, y
  printf("%s", essq);
  return 0;
}
//------------------------ End gotoxy ----------------------------------

//+++++++++++++++++++++++ Start clrscr ++++++++++++++++++++++++++
void clrscr(int StartRow, int EndRow) {
  int i, i2;

  if (EndRow < StartRow){
    i = EndRow;
    EndRow = StartRow;
    StartRow = i;
  }
  gotoxy(1, StartRow);
  for (i = 0; i <= EndRow - StartRow; i++){
    for(i2 = 0; i2 < MaxCols; i2++){
      printf(" ");
    }
    printf("\n");
  }
}
//----------------------- End clrscr ----------------------------

//+++++++++++++++++++++++ Start kbhit ++++++++++++++++++++++++++++++++++
//Thanks to Undertech Blog, http://www.undertec.de/blog/2009/05/kbhit_und_getch_fur_linux.html
int kbhit(void) {

   struct termios term, oterm;
   int fd = 0;
   int c = 0;

   tcgetattr(fd, &oterm);
   memcpy(&term, &oterm, sizeof(term));
   term.c_lflag = term.c_lflag & (!ICANON);
   term.c_cc[VMIN] = 0;
   term.c_cc[VTIME] = waitForKeyHit;
   tcsetattr(fd, TCSANOW, &term);
   c = getchar();
   tcsetattr(fd, TCSANOW, &oterm);
   if (c != -1)
   ungetc(c, stdin);

   return ((c != -1) ? 1 : 0);

}
//------------------------ End kbhit -----------------------------------

//+++++++++++++++++++++++ Start getch ++++++++++++++++++++++++++++++++++
//Thanks to Undertech Blog, http://www.undertec.de/blog/2009/05/kbhit_und_getch_fur_linux.html
int getch(){
   static int ch = -1, fd = 0;
   struct termios new, old;

   fd = fileno(stdin);
   tcgetattr(fd, &old);
   new = old;
   new.c_lflag &= ~(ICANON|ECHO);
   tcsetattr(fd, TCSANOW, &new);
   ch = getchar();
   tcsetattr(fd, TCSANOW, &old);

//   printf("ch=%d ", ch);

   return ch;
}
//------------------------ End getch -----------------------------------

//++++++++++++++++++++++ Start MessageText +++++++++++++++++++++++++++++
void MessageText(char *message, int x, int y, int alignment){
  int i;
  char TextLine[300];

  clrscr(y, y);
  gotoxy (x, y);

  TextLine[0] = '\0';
  if(alignment == 1){
    for(i=0; i < (MaxCols - strlen(message)) / 2 ; i++){
      strcat(TextLine, " ");
    }
  }
  strcat(TextLine, message);

  printf("%s\n", TextLine);
}
//-------------------------- End MessageText ---------------------------

//++++++++++++++++++++++ Start PrintRow ++++++++++++++++++++++++++++++++
void PrintRow(char character, int y){
  int i;
  gotoxy (1, y);
  for(i=0; i<MaxCols;i++){
    printf("%c", character);
  }
}
//-------------------------- End PrintRow ------------------------------

//+++++++++++++++++++++++++ ErrorText +++++++++++++++++++++++++++++
void ErrorText(char *message){
  clrscr(MessageY + 2, MessageY + 2);
  gotoxy (1, MessageY + 2);
  printf("Last error: %s", message);
}
//----------------------------- ErrorText ---------------------------

//+++++++++++++++++++++++++ PrintMenue_01 ++++++++++++++++++++++++++++++
void PrintMenue_01(){
  char TextLine[300];

   clrscr(1, MessageY-2);
   MessageText("*** Main menu plotter ***", 1, 1, 1);
   sprintf(TextLine, "M            - toggle move length, current value = %0.3lf mm", MoveLength);
   MessageText(TextLine, 10, 2, 0);
   MessageText("Cursor right - move plotter in positive X direction", 10, 3, 0);
   MessageText("Cursor left  - move plotter in negative X direction", 10, 4, 0);
   MessageText("Cursor up    - move plotter in positive Y direction", 10, 5, 0);
   MessageText("Cursor down  - move plotter in negative Y direction", 10, 6, 0);
   MessageText("Page up      - lift miller", 10, 7, 0);
   MessageText("Page down    - lower miller", 10, 8, 0);
   if(relayOn == 0){
     sprintf(TextLine, "R            - Relay ON/OFF, currently: OFF");
   }
   else{
     sprintf(TextLine, "R            - Relay ON/OFF, currently: ON");
   }
   MessageText(TextLine, 10, 9, 0);
   sprintf(TextLine, "F            - choose file: \"%s\", W=%0.1lfmm, H=%0.1lfmm", FileName, xMax - xMin, yMax-yMin);
   MessageText(TextLine, 10, 10, 0);
   MessageText("P            - process file", 10, 11, 0);

   MessageText("Esc          - leave program", 10, 13, 0);

}
//------------------------- PrintMenue_01 ------------------------------

//+++++++++++++++++++++++++ PrintMenue_02 ++++++++++++++++++++++++++++++
char *PrintMenue_02(int StartRow, int selected){
  char TextLine[300];
  char OpenDirName[1000];
  static char FileName[101];
  DIR *pDIR;
  struct dirent *pDirEnt;
  int i = 0;
  int Discard = 0;

  clrscr(1, MessageY-2);
  MessageText("*** Choose plotter file ***", 1, 1, 1);

  strcpy(OpenDirName, PicturePath);


  pDIR = opendir(OpenDirName);
  if ( pDIR == NULL ) {
    sprintf(TextLine, "Could not open directory '%s'!", OpenDirName);
    MessageText(TextLine, 1, 4, 1);
    getch();
    return( "" );
  }

  FilesFound = 0;
  pDirEnt = readdir( pDIR );
  while ( pDirEnt != NULL && i < MaxFileRows) {
    //printf("%s\n", pDirEnt->d_name);
    if(strlen(pDirEnt->d_name) > 3){
      if(memcmp(pDirEnt->d_name + strlen(pDirEnt->d_name)-6, ".gcode",6) == 0 || memcmp(pDirEnt->d_name + strlen(pDirEnt->d_name)-4, ".ngc",4) == 0 || memcmp(pDirEnt->d_name + strlen(pDirEnt->d_name)-3, ".nc",3) == 0){
        FilesFound++;
        if(Discard >= StartRow){
          if(i + StartRow == selected){
            sprintf(TextLine, ">%s<", pDirEnt->d_name);
            strcpy(FileName, pDirEnt->d_name);
          }
          else{
            sprintf(TextLine, " %s ", pDirEnt->d_name);
          }
          MessageText(TextLine, 1, 3 + i, 0);
          i++;
        }
        Discard++;
      }
    }
    pDirEnt = readdir( pDIR );
  }

  gotoxy(MessageX, MessageY + 1);
  printf("Choose file using up/down keys and confirm with 'Enter' or press 'Esc' to cancel.");


  return (FileName);
}
//------------------------- PrintMenue_02 ------------------------------


//+++++++++++++++++++++++++ PrintMenue_03 ++++++++++++++++++++++++++++++
void PrintMenue_03(char *FullFileName, long CurrentLine, long NumberOfLines, char *gcodeLine, long StartTime){
  char TextLine[300];
  long CurrentTime, ProcessHours = 0, ProcessMinutes = 0, ProcessSeconds = 0;

   CurrentTime = time(0);

   CurrentTime -= StartTime;

   while (CurrentTime > 3600){
     ProcessHours++;
     CurrentTime -= 3600;
   }
   while (CurrentTime > 60){
     ProcessMinutes++;
     CurrentTime -= 60;
   }
   ProcessSeconds = CurrentTime;

   clrscr(1, MessageY - 2);
   MessageText("*** Plotting file ***", 1, 1, 1);

   sprintf(TextLine, "File name: %s", FullFileName);
   MessageText(TextLine, 10, 3, 0);
   sprintf(TextLine, "Line number / Total number of lines: %ld / %ld", CurrentLine, NumberOfLines);
   MessageText(TextLine, 10, 4, 0);
   sprintf(TextLine, "Current g code command: %s    ", gcodeLine);
   MessageText(TextLine, 10, 5, 0);
   sprintf(TextLine, "Process time: %02ld:%02ld:%02ld", ProcessHours, ProcessMinutes, ProcessSeconds);
   MessageText(TextLine, 10, 6, 0);


}
//------------------------- PrintMenue_03 ------------------------------


//+++++++++++++++++++++++ Start readport ++++++++++++++++++++++++++
void  clearReadPort(void){
  int n;
  char buff;
  while (read(fd, &buff, 1) > 0);
}
//------------------------ End readport ----------------------------------

//+++++++++++++++++++++++ Start readport ++++++++++++++++++++++++++
char  readport(void){
  int n;
    char buff;
  n = read(fd, &buff, 1);
  if(n > 0){
    return buff;
  }
  return 0;
}
//------------------------ End readport ----------------------------------

//+++++++++++++++++++++++ Start sendport ++++++++++++++++++++++++++
void sendport(char *CommandLineToSend){
  int n;
  int BytesInBuffer;
  int CommandLength;


//  ioctl(fd,TIOCOUTQ,&BytesInBuffer);
//  while(BytesInBuffer > MAX_ARDUINO_BUFFER){//Wait until bytes in write buffer are below maximum
//    ioctl(fd,TIOCOUTQ,&BytesInBuffer);
//  }


  clearReadPort();


  CommandLength = strlen(CommandLineToSend);
  if(CommandLineToSend[CommandLength] != 13){
    CommandLineToSend[CommandLength] = 13;
    CommandLineToSend[CommandLength + 1] = '\0';
    CommandLength++;
  }
  n = write(fd, CommandLineToSend, CommandLength);

  if (n < 0){
    ErrorText("write() of value failed!\r");
  }
  else{
//    arduinoBuffer++;
//    if(arduinoBuffer == MAX_ARDUINO_BUFFER){
      while(readport() == 0);//Wait for any response from grbl
//      arduinoBuffer = 0;
//    }
  }




}
//------------------------ End sendport ----------------------------------

//+++++++++++++++++++++++ Start openport ++++++++++++++++++++++++++
void openport(void){

    fd = open(ARDUINOPORT, O_RDWR | O_NOCTTY | O_NDELAY);
    if (fd == -1)  {
      ErrorText("init_serialport: Unable to open port ");
    }

    if (tcgetattr(fd, &TermOpt) < 0) {
      ErrorText("init_serialport: Couldn't get term attributes");
    }
    speed_t brate = BAUDRATE; // let you override switch below if needed

    cfsetispeed(&TermOpt, brate);
    cfsetospeed(&TermOpt, brate);

    // 8N1
    TermOpt.c_cflag &= ~PARENB;
    TermOpt.c_cflag &= ~CSTOPB;
    TermOpt.c_cflag &= ~CSIZE;
    TermOpt.c_cflag |= CS8;
    // no flow control
    TermOpt.c_cflag &= ~CRTSCTS;

    TermOpt.c_cflag |= CREAD | CLOCAL;  // turn on READ & ignore ctrl lines
    TermOpt.c_iflag &= ~(IXON | IXOFF | IXANY); // turn off s/w flow ctrl

    TermOpt.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG); // make raw
    TermOpt.c_oflag &= ~OPOST; // make raw

    // see: http://unixwiz.net/techtips/termios-vmin-vtime.html
    TermOpt.c_cc[VMIN]  = 0;
    TermOpt.c_cc[VTIME] = 20;

    if( tcsetattr(fd, TCSANOW, &TermOpt) < 0) {
      ErrorText("init_serialport: Couldn't set term attributes");
    }

}
//------------------------ End openport ----------------------------------

//+++++++++++++++++++++++++++++ get next Gcode command set ++++++++
long GetGcodeCoordinates(int mirrorX, int mirrorY, int mirrorZ){
  char TextLine[1000];
  long i;
  char a = 0;
  char GCodeCharacter;
  int EndOfLine = 0;
  double GCodeNumber = 0.0;
  char *pEnd;


  a=0;
  i=0;
  TextLine[0] = '\0';
  EndOfLine = 0;
  while(!(feof(GcodeFile)) && EndOfLine == 0){
    if(a == '1' || a == '2' || a == '3' || a == '4' || a == '5' || a == '6' || a == '7' || a == '8' || a == '9' || a == '0' || a == '.' || a == '-'){
      TextLine[i] = a;
      TextLine[i+1] = '\0';
      i++;
      if(i==900){
        ErrorText("Line in Gcode file too long!");
      }
    }
    else{
      if(i > 0){
        LastGcodeNumber = GCodeNumber;
        GCodeNumber = strtod(TextLine, &pEnd);
        if(GCodeCharacter == 'x' || GCodeCharacter == 'X'){
          newX = GCodeNumber;
          if (mirrorX == 1){
            newX = xMax - newX;
          }
        }
        if(GCodeCharacter == 'y' || GCodeCharacter == 'Y'){
          newY = GCodeNumber;
          if (mirrorY == 1){
            newY = yMax - newY;
          }
        }
        if(GCodeCharacter == 'z' || GCodeCharacter == 'Z'){
          newZ = GCodeNumber;
        }
//        printf("GCodeNumber=%lf, GCodeCharacter='%c', TextLine = %s\n", GCodeNumber, GCodeCharacter, TextLine);// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        if(GCodeCharacter == 'G' || GCodeCharacter == 'g' && GCodeNumber == 90){
          AbsolutePositioning = 1;
        }

        if(GCodeCharacter == 'G' || GCodeCharacter == 'g' && GCodeNumber == 91){
          AbsolutePositioning = 0;
        }

        if(GCodeCharacter == 'G' || GCodeCharacter == 'g' && GCodeNumber == 0){
          feedRateNew = MAX_FEED_RATE;
        }

        if(GCodeCharacter == 'G' || GCodeCharacter == 'g' && GCodeNumber > 0 && GCodeNumber < 4){
          feedRateNew = latestFeedRate;
        }

        if(GCodeCharacter == 'F' || GCodeCharacter == 'f' && GCodeNumber > 0){
          feedRateNew = GCodeNumber;
          if(feedRateNew > MAX_FEED_RATE){
            feedRateNew = MAX_FEED_RATE;
          }
          latestFeedRate = feedRateNew;
        }

        if(GCodeCharacter == 'M' || GCodeCharacter == 'm' && GCodeNumber == 3){
          SpindleOn = 1;
        }
        if(GCodeCharacter == 'M' || GCodeCharacter == 'm' && GCodeNumber == 5){
          SpindleOn = 0;
        }

        if(GCodeCharacter == 'P' || GCodeCharacter == 'p'){
          if(LastGCodeCharacter == 'G' || LastGCodeCharacter == 'g' && LastGcodeNumber == 4){
            PauseTime = GCodeNumber;
          }
        }

      }
      i = 0;
      if(a != ' ' && a != ';' && a != 10 && a != 9 && a != 0){
        LastGCodeCharacter = GCodeCharacter;
        GCodeCharacter = a;
      }
    }
    if(a == ';'){//Discard until end of line
      while(!(feof(GcodeFile)) && a != 10){
        fread(&a, 1, 1, GcodeFile);
      }
    }
    if(a == 10){//End of line
      EndOfLine = 1;
    }
    else{
      fread(&a, 1, 1, GcodeFile);
    }
  }

  if(feof(GcodeFile)){
    return 0;
  }
  return 1;
}

//----------------------------- get next Gcode command set --------



//+++++++++++++++++++++++++++++ get next line of Gcode commands ++++++++
int GetGcodeLine(char *TextLine){
  long i;
  char a = 0;
  char GCodeCharacter;
  int EndOfLine = 0;
  double GCodeNumber = 0.0;
  char *pEnd;


  a=0;
  i=0;
  TextLine[0] = '\0';
  EndOfLine = 0;
  while(!(feof(GcodeFile)) && EndOfLine == 0){
    if(a == ';'){//Discard until end of line
      while(!(feof(GcodeFile)) && a != 10){
        fread(&a, 1, 1, GcodeFile);
      }
    }
    if(!(a == 13 || a == 10 || a == 0 || a == ';')){
      TextLine[i] = a;
      TextLine[i+1] = '\0';
      i++;
      if(i==900){
        ErrorText("Line in Gcode file too long!");
      }
    }
    if(a == 10 || a == 13){//End of line
      EndOfLine = 1;
    }
    else{
      fread(&a, 1, 1, GcodeFile);
    }
  }

  if(feof(GcodeFile)){
    return 0;
  }
  return 1;
}

//----------------------------- get next line of Gcode commands --------

//++++++++++++++++++++++++++++++++++++++ KeyControl ++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void KeyControl(int action){
int i;
char textLine[1000];

  i = 0;
  SingleKey = 1;
  KeyCode[0] = 0;
  KeyCode[1] = 0;
  KeyCode[2] = 0;
  KeyCode[3] = 0;
  KeyCode[4] = 0;
  KeyHit = 0;
  while (kbhit()){
    KeyHit = getch();
    KeyCode[i] = KeyHit;
    i++;
    if(i == 5){
      i = 0;
    }
    if(i > 1){
      SingleKey = 0;
    }
  }
  if(SingleKey == 0){
    KeyHit = 0;
  }

  if(action == 1 || action == 2){
    //Move X-axis
    if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 68 && KeyCode[3] == 0 && KeyCode[4] == 0){
	  currentX -= MoveLength;
      sprintf(textLine, "G0 X%0.3lf", currentX);
      sendport(textLine);
    }

    if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 67 && KeyCode[3] == 0 && KeyCode[4] == 0){
	  currentX += MoveLength;
      sprintf(textLine, "G0 X%0.3lf", currentX);
      sendport(textLine);
    }

    //Move Y-axis
    if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 65 && KeyCode[3] == 0 && KeyCode[4] == 0){
	  currentY += MoveLength;
      sprintf(textLine, "G0 Y%0.3lf", currentY);
      sendport(textLine);
    }

    if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 66 && KeyCode[3] == 0 && KeyCode[4] == 0){
	  currentY -= MoveLength;
      sprintf(textLine, "G0 Y%0.3lf", currentY);
      sendport(textLine);
    }

    //Pen UP/DOWN
    if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 53 && KeyCode[3] == 126 && KeyCode[4] == 0){
	  currentZ += MoveLength;
      sprintf(textLine, "G0 Z%0.3lf", currentZ);
      sendport(textLine);
    }

    if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 54 && KeyCode[3] == 126 && KeyCode[4] == 0){
	  currentZ -= MoveLength;
      sprintf(textLine, "G0 Z%0.3lf", currentZ);
      sendport(textLine);
    }



    if(KeyHit == 'm'){
      MoveLength *= 10.0;
      if(MoveLength == 100.0){
        MoveLength = 0.1;
      }

      PrintMenue_01();
    }


    if(KeyHit == 'r'){
      if(relayOn == 1){
		sprintf(textLine, "M05");
        sendport(textLine);//Turn relay off
        relayOn = 0;
      }
      else{
		sprintf(textLine, "M03");
        sendport(textLine);//Turn relay on
        relayOn = 1;
      }
      PrintMenue_01();
    }

  }


}
//-------------------------------------- KeyControl --------------------------------------------------------


//######################################################################
//################## Main ##############################################
//######################################################################

int main(int argc, char **argv){

  int MenueLevel = 0;
  char FullFileName[1000] = "";
  char FileNameOld[1000] = "";
  struct winsize terminal;
  int i;
  int FileSelected = 0;
  int FileStartRow = 0;
  char *pEnd;
  char TextLine[30000];
  long coordinateCount = 0;
  long maxCoordinateCount = 0;
  char a;
  int ReadState = 0;
  long PlotStartTime = 0;


  openport();

  printf("\n\nWaiting for response from CNC (machine pluged in USB port?)...\n");

  //Wait for ']' from Arduino
  while(readport() != ']');


  StepPause = 1.0 / (StepsPermmX * MAX_FEED_RATE  / 60.0) * 1000.0;

  sprintf(TextLine, "G92 X0 Y0 Z0");
  sendport(TextLine);//Reset grbl coordinates to zero


  strcpy(FileName, "noFiLE");

  getcwd(PicturePath, 1000);
  strcat(PicturePath, "/files");
  printf("PicturePath=>%s<", PicturePath);

  if(ioctl(STDOUT_FILENO, TIOCGWINSZ, &terminal)<0){
    printf("Can't get size of terminal window");
  }
  else{
    MaxRows = terminal.ws_row;
    MaxCols = terminal.ws_col;
    MessageY = MaxRows-3;
  }

  clrscr(1, MaxRows);
  PrintRow('-', MessageY - 1);
  PrintMenue_01();

  MaxFileRows = MaxRows - 10;

  while (1){
    MessageText("Waiting for key press.", MessageX, MessageY, 0);

    if(MenueLevel == 0){
      KeyControl(1);

      if(KeyHit == 'f'){
        strcpy(FileNameOld, FileName);
        strcpy(FileName, PrintMenue_02(FileStartRow, FileSelected));
        MenueLevel = 1;
      }

      if(KeyHit == 's'){
        OldScale = Scale;
        MessageText("Type new scale value: ", 1, MessageY, 0);
        gotoxy(23, MessageY);
        scanf("%lf", &Scale);
        if(Scale == 0){
          Scale = OldScale;
        }
        else{
          PrintMenue_01();
        }
      }





      if(KeyHit == 'p'){//Plot file
        //MessageText("13 seconds until plotting starts !!!!!!!!!!!!!!!!!", 1, 20, 0);
        //sleep(10);
        //MessageText("3 seconds until plotting starts !!!!!!!!!!!!!!!!!", 1, 20, 0);
        //sleep(3);
        if(kbhit()){
          getch();
          KeyHit = 27;
        }
        if(strcmp(FileName, "noFiLE") != 0){
          if((GcodeFile=fopen(FullFileName,"rb"))==NULL){
            sprintf(TextLine, "Can't open file '%s'!\n", FullFileName);
            strcpy(FileName, "NoFiLE");
            ErrorText(TextLine);
          }
        }
        if(strcmp(FileName, "noFiLE") != 0 && KeyHit != 27){
          char TextLine[1000];

		  sprintf(TextLine, "G92 X0 Y0 Z0");
		  sendport(TextLine);//Reset grbl coordinates to zero

          PlotStartTime = time(0);
          stopPlot = 0;
          while (stopPlot == 0){
			if(GetGcodeLine(TextLine) == 0){
			  stopPlot = 1;
			}
			else{
              sendport(TextLine);
		    }
            coordinateCount++;
            PrintMenue_03(FullFileName, coordinateCount, maxCoordinateCount, TextLine, PlotStartTime);
  		    waitForKeyHit = 0;
		    if(kbhit()){
			  if(getch() == 27){
			    stopPlot = 1;
				TextLine[0] = 24; TextLine[1] = '\0';//Send soft-reset
				sendport(TextLine);
				sleep(1);
				close(fd);
				sleep(1);
				openport();
				//Wait for ']' from Arduino
				while(readport() != ']');
				sprintf(TextLine, "G92 X0 Y0 Z0");
				sendport(TextLine);//Reset grbl coordinates to zero
				currentX = 0.0;				
				currentY = 0.0;				
				currentZ = 0.0;				
				relayOn = 0;
			  }
		    }
		    waitForKeyHit = 1;
          }

          fclose(GcodeFile);
          stopPlot = 0;
          PrintMenue_03(FullFileName, coordinateCount, maxCoordinateCount, TextLine, PlotStartTime);
          while(kbhit()){
            getch();
          }

		  sprintf(TextLine, "M05");
		  sendport(TextLine);//Turn relay off
		  relayOn = 0;

          MessageText("Finished! Press any key to return to main menu.", MessageX, MessageY, 0);
          getch();
          PrintMenue_01();

        }//if(strcmp(FileName, "noFiLE") != 0){
      }//if(KeyHit == 'p'){


    }//if(MenueLevel == 0){

    if(MenueLevel == 1){//Select file
      KeyControl(0);
      if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 66 && KeyCode[3] == 0 && KeyCode[4] == 0){
        if(FileSelected < FilesFound - 1){
          FileSelected++;
          if(FileSelected > MaxFileRows - 2){
            FileStartRow = FileSelected - MaxFileRows + 2;
          }
          strcpy(FileName, PrintMenue_02(FileStartRow, FileSelected));
        }
      }

      if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 65 && KeyCode[3] == 0 && KeyCode[4] == 0){
        if(FileSelected > 0){
          if(FileSelected == FileStartRow + 1){
            if(FileStartRow > 0){
              FileStartRow--;
            }
          }
          FileSelected--;
          strcpy(FileName, PrintMenue_02(FileStartRow, FileSelected));
        }
      }

      if(KeyHit == 10){//Read file and store values
        MenueLevel = 0;
        clrscr(MessageY + 1, MessageY + 1);
        strcpy(FullFileName, PicturePath);
        strcat(FullFileName, "/");
        strcat(FullFileName, FileName);
        if((GcodeFile=fopen(FullFileName,"rb"))==NULL){
          sprintf(TextLine, "Can't open file '%s'!\n", FullFileName);
          ErrorText(TextLine);
          strcpy(FileName, "NoFiLE");
        }
        else{
          //printf("Opened file '%s'!\n", FullFileName);

          xMin=1000000;
          xMax=-1000000;
          yMin=1000000;
          yMax=-1000000;
          zMin=1000000;
          zMax=-1000000;

          newX = -99999999;
          newY = -99999999;
          newZ = -99999999;
          maxCoordinateCount = 0;

          while (GetGcodeCoordinates(0, 0, 0) > 0){

            if(newX > xMax && newX != -99999999){
              xMax = newX;
            }
            if(newX < xMin && newX != -99999999){
              xMin = newX;
            }

            if(newY > yMax && newY != -99999999){
              yMax = newY;
            }
            if(newY < yMin && newY != -99999999){
              yMin = newY;
            }

            if(newZ > zMax && newZ != -99999999){
              zMax = newZ;
            }
            if(newZ < zMin && newZ != -99999999){
              zMin = newZ;
            }

            maxCoordinateCount++;
          }

          fclose(GcodeFile);
          Scale = 1.0;
          //printf("xMax = %ld, xMin=%ld, yMax = %ld, yMin = %ld  ", xMax, xMin, yMax, yMin);
          //getch();
        }
        PrintMenue_01();
      }//if(KeyHit == 10){

    }//if(MenueLevel == 1){


    if(KeyHit == 27){
      if(MenueLevel == 0){
        clrscr(MessageY + 1, MessageY + 1);
        MessageText("Exit program (y/n)?", MessageX, MessageY + 1, 0);
        while(KeyHit != 'y' && KeyHit != 'n'){
          KeyHit = getch();
          if(KeyHit == 'y'){
            exit(0);
          }
        }
      }
      if(MenueLevel == 1){
        MenueLevel = 0;
        strcpy(FileName, FileNameOld);
        PrintMenue_01();
      }
      clrscr(MessageY + 1, MessageY + 1);
    }
  }

  return 0;
}
